/*
 * QEstEidCommon
 *
 * Copyright (C) 2009-2012 Jargo Kõster <jargo@innovaatik.ee>
 * Copyright (C) 2009-2012 Raul Metsma <raul@innovaatik.ee>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include "Common.h"

#include "SslCertificate.h"
#include "TokenData.h"
#include "Settings.h"

#include <QtCore/QDebug>
#include <QtCore/QDir>
#include <QtCore/QFileInfo>
#include <QtCore/QProcess>
#include <QtCore/QTextStream>
#include <QtCore/QUrl>
#include <QtGui/QDesktopServices>
#include <QtGui/QLabel>
#include <QtGui/QPalette>
#include <QtGui/QTextDocument>

#include <stdlib.h>

#if defined(Q_OS_WIN)
#include <QtCore/QLibrary>
#include <QtCore/QSettings>

#include <qt_windows.h>
#include <mapi.h>
#include <Shellapi.h>
#elif defined(Q_OS_MAC)
#include <QtCore/QXmlStreamReader>
#include <Carbon/Carbon.h>
#include <sys/utsname.h>
#ifdef USE_SECURITY
#include <Security/Authorization.h>
#endif
#endif

static QString packageName( const QString &name, const QString &ver, bool withName )
{ return withName ? name + " (" + ver + ")" : ver; }

Common::Common( int &argc, char **argv )
	: BaseApplication( argc, argv )
{
	Q_INIT_RESOURCE(common_images);
	Q_INIT_RESOURCE(common_tr);
#if defined(Q_OS_WIN) || defined(Q_OS_MAC)
	setLibraryPaths( QStringList() << applicationDirPath() );
#endif
	setStyleSheet(
		"QDialogButtonBox { dialogbuttonbox-buttons-have-icons: 0; }\n" );
	QPalette p = palette();
	p.setBrush( QPalette::Link, QBrush( "#509B00" ) );
	p.setBrush( QPalette::LinkVisited, QBrush( "#509B00" ) );
	setPalette( p );

	qRegisterMetaType<QSslCertificate>("QSslCertificate");
	qRegisterMetaType<TokenData>("TokenData");

	QDesktopServices::setUrlHandler( "browse", this, "browse" );
	QDesktopServices::setUrlHandler( "mailto", this, "mailTo" );

#if defined(Q_OS_WIN)
	AllowSetForegroundWindow( ASFW_ANY );
#elif defined(Q_OS_MAC)
	macEvents = 0;
#endif
}

Common::~Common()
{
#if defined(Q_OS_MAC)
	deinitMacEvents();
#endif
}

QString Common::applicationOs()
{
#if defined(Q_OS_LINUX)
	QProcess p;
	p.start( "lsb_release", QStringList() << "-s" << "-d" );
	p.waitForFinished();
	return QString::fromLocal8Bit( p.readAll().trimmed() );
#elif defined(Q_OS_MAC)
	struct utsname unameData;
	uname(&unameData);
	SInt32 major, minor, bugfix;
	if( Gestalt(gestaltSystemVersionMajor, &major) == noErr &&
			Gestalt(gestaltSystemVersionMinor, &minor) == noErr &&
			Gestalt(gestaltSystemVersionBugFix, &bugfix) == noErr )
		return QString( "Mac OS %1.%2.%3 (%4/%5)" )
			.arg( major ).arg( minor ).arg( bugfix ).arg( QSysInfo::WordSize ).arg( unameData.machine );
#elif defined(Q_OS_WIN)
	OSVERSIONINFOEX osvi = { sizeof( OSVERSIONINFOEX ) };
	if( GetVersionEx( (OSVERSIONINFO *)&osvi ) )
	{
		SYSTEM_INFO si;
		typedef void (WINAPI *PGNSI)(LPSYSTEM_INFO);
		if( PGNSI pGNSI = PGNSI( QLibrary( "kernel32" ).resolve( "GetNativeSystemInfo" ) ) )
			pGNSI( &si );
		else
			GetSystemInfo( &si );
		QString os;
		switch( osvi.dwMajorVersion )
		{
		case 5:
			switch( osvi.dwMinorVersion )
			{
			case 0:
				os = QString( "Windows 2000 %1" ).arg( osvi.wProductType == VER_NT_WORKSTATION ? "Professional" : "Server" );
				break;
			case 1:
				os = QString( "Windows XP %1" ).arg( osvi.wSuiteMask & VER_SUITE_PERSONAL ? "Home" : "Professional" );
				break;
			case 2:
				if( GetSystemMetrics( SM_SERVERR2 ) )
					os = "Windows Server 2003 R2";
				else if( osvi.wProductType == VER_NT_WORKSTATION && si.wProcessorArchitecture == PROCESSOR_ARCHITECTURE_AMD64 )
					os = "Windows XP Professional";
				else
					os = "Windows Server 2003";
				break;
			default: break;
			}
			break;
		case 6:
			switch( osvi.dwMinorVersion )
			{
			case 0:
				os = osvi.wProductType == VER_NT_WORKSTATION ? "Windows Vista" : "Windows Server 2008";
				break;
			case 1:
				os = osvi.wProductType == VER_NT_WORKSTATION ? "Windows 7" : "Windows Server 2008 R2";
				break;
			case 2:
				os = osvi.wProductType == VER_NT_WORKSTATION ? "Windows 8" : "Windows Server 8";
				break;
			default: break;
			}
			break;
		default: break;
		}
		if( !os.isEmpty() )
		{
			if( osvi.szCSDVersion > 0 )
				os += " " + QString( (const QChar*)osvi.szCSDVersion );

			return QString( "%1 (%2 bit)" ).arg( os )
				.arg( si.wProcessorArchitecture == PROCESSOR_ARCHITECTURE_AMD64 ? "64" : "32" );
		}
	}
	else
	{
		switch( QSysInfo::WindowsVersion )
		{
		case QSysInfo::WV_2000: return "Windows 2000"; break;
		case QSysInfo::WV_XP: return "Windows XP"; break;
		case QSysInfo::WV_2003: return "Windows 2003"; break;
		case QSysInfo::WV_VISTA: return "Windows Vista"; break;
		case QSysInfo::WV_WINDOWS7: return "Windows 7"; break;
		default: break;
		}
	}
#endif

	return tr("Unknown OS");
}

quint8 Common::cardsOrderScore( QChar c )
{
	switch( c.toAscii() )
	{
	case 'N': return 6;
	case 'A': return 5;
	case 'P': return 4;
	case 'E': return 3;
	case 'F': return 2;
	case 'B': return 1;
	default: return 0;
	}
}

bool Common::cardsOrder( const QString &s1, const QString &s2 )
{
	QRegExp r("(\\w{1,2})(\\d{7})");
	if( r.indexIn( s1 ) == -1 )
		return false;
	QStringList cap1 = r.capturedTexts();
	if( r.indexIn( s2 ) == -1 )
		return false;
	QStringList cap2 = r.capturedTexts();
	// new cards to front
	if( cap1[1].size() != cap2[1].size() )
		return cap1[1].size() > cap2[1].size();
	// card type order
	if( cap1[1][0] != cap2[1][0] )
		return cardsOrderScore( cap1[1][0] ) > cardsOrderScore( cap2[1][0] );
	// card version order
	if( cap1[1].size() > 1 && cap2[1].size() > 1 && cap1[1][1] != cap2[1][1] )
		return cap1[1][1] > cap2[1][1];
	// serial number order
	return cap1[2].toUInt() > cap2[2].toUInt();
}

void Common::browse( const QUrl &url )
{
	QUrl u = url;
	u.setScheme( "file" );
#if defined(Q_OS_WIN)
	if( QProcess::startDetached( "explorer", QStringList() << "/select," <<
		QDir::toNativeSeparators( u.toLocalFile() ) ) )
		return;
#elif defined(Q_OS_MAC)
	if( QProcess::startDetached( "open", QStringList() << "-R" << u.toLocalFile() ) )
		return;
#endif
	QDesktopServices::openUrl( QUrl::fromLocalFile( QFileInfo( u.toLocalFile() ).absolutePath() ) );
}

QUrl Common::helpUrl()
{
	QString lang = Settings::language();
	QUrl u( "http://support.sk.ee" );
	if( lang == "en" ) u = "http://support.sk.ee/eng/";
	if( lang == "ru" ) u = "http://support.sk.ee/ru/";
	u.addQueryItem( "app", applicationName() );
	u.addQueryItem( "appver", applicationVersion() );
	u.addQueryItem( "os", applicationOs() );
	return u;
}

bool Common::event( QEvent *e )
{
#ifdef Q_OS_MAC
	// Load here because cocoa NSApplication overides events
	if( e->type() == QEvent::ApplicationActivate && !macEvents )
		initMacEvents();
#endif
	return BaseApplication::event( e );
}

QString Common::fileSize( quint64 bytes )
{
	const quint64 kb = 1024;
	const quint64 mb = 1024 * kb;
	const quint64 gb = 1024 * mb;
	const quint64 tb = 1024 * gb;
	if( bytes >= tb )
		return QString( "%1 TB" ).arg( qreal(bytes) / tb, 0, 'f', 3 );
	if( bytes >= gb )
		return QString( "%1 GB" ).arg( qreal(bytes) / gb, 0, 'f', 2 );
	if( bytes >= mb )
		return QString( "%1 MB" ).arg( qreal(bytes) / mb, 0, 'f', 1 );
	if( bytes >= kb )
		return QString( "%1 KB" ).arg( bytes / kb );
	return QString( "%1 B" ).arg( bytes );
}

void Common::mailTo( const QUrl &url )
{
#if defined(Q_OS_WIN)
	QString mail = QSettings( "HKEY_CURRENT_USER\\Software\\Clients\\Mail",
		QSettings::NativeFormat ).value( "." ).toString();
	bool utf8 = QSettings( "HKEY_LOCAL_MACHINE\\Software\\Clients\\Mail\\" + mail,
		QSettings::NativeFormat ).value( "SupportUTF8", false ).toBool();

	QString file = url.queryItemValue( "attachment" );
	QByteArray filePath = QDir::toNativeSeparators( file ).toLocal8Bit();
	QByteArray fileName = QFileInfo( file ).fileName().toLocal8Bit();
	QByteArray subject = url.queryItemValue( "subject" ).toLocal8Bit();

	MapiFileDesc doc[1];
	doc[0].ulReserved = 0; //utf8 ? CP_UTF8 : 0;
	doc[0].flFlags = 0;
	doc[0].nPosition = -1;
	doc[0].lpszPathName = const_cast<char*>(filePath.constData());
	doc[0].lpszFileName = const_cast<char*>(fileName.constData());
	doc[0].lpFileType = 0;

	// Create message
	MapiMessage message;
	message.ulReserved = 0; //utf8 ? CP_UTF8 : 0;
	message.lpszSubject = const_cast<char*>(subject.constData());
	message.lpszNoteText = "";
	message.lpszMessageType = 0;
	message.lpszDateReceived = 0;
	message.lpszConversationID = 0;
	message.flFlags = 0;
	message.lpOriginator = 0;
	message.nRecipCount = 0;
	message.lpRecips = 0;
	message.nFileCount = 1;
	message.lpFiles = lpMapiFileDesc(&doc);

	QLibrary lib("mapi32");
	if( LPMAPISENDMAIL mapi = LPMAPISENDMAIL(lib.resolve("MAPISendMail")) )
	{
		switch( mapi( NULL, 0, &message, MAPI_LOGON_UI|MAPI_DIALOG, 0 ) )
		{
		case SUCCESS_SUCCESS:
		case MAPI_E_USER_ABORT:
		case MAPI_E_LOGIN_FAILURE:
			return;
		default: break;
		}
	}
#elif defined(Q_OS_MAC)
	CFURLRef emailUrl = CFURLCreateWithString( kCFAllocatorDefault, CFSTR("mailto:"), 0 );
	CFURLRef appUrl = 0;
	CFStringRef appPath = 0;
	if( LSGetApplicationForURL( emailUrl, kLSRolesAll, NULL, &appUrl ) == noErr )
	{
		appPath = CFURLCopyFileSystemPath( appUrl, kCFURLPOSIXPathStyle );
		CFRelease( appUrl );
	}
	CFRelease( emailUrl );

	if( appPath )
	{
		QProcess p;
		p.start( "/usr/bin/osascript", QStringList() << "-" << url.queryItemValue("attachment") << url.queryItemValue("subject") );
		p.waitForStarted();
		QTextStream s( &p );
		if( CFStringCompare( appPath, CFSTR("/Applications/Mail.app"), 0 ) == kCFCompareEqualTo )
		{
			s << "on run argv" << endl
				<< "set vattachment to (item 1 of argv)" << endl
				<< "set vsubject to (item 2 of argv)" << endl
				<< "tell application \"Mail\"" << endl
				<< "set composeMessage to make new outgoing message at beginning with properties {visible:true}" << endl
				<< "tell composeMessage" << endl
				<< "set subject to vsubject" << endl
				<< "set content to \" \"" << endl
				<< "tell content" << endl
				<< "make new attachment with properties {file name: vattachment} at after the last word of the last paragraph" << endl
				<< "end tell" << endl
				<< "end tell" << endl
				<< "activate" << endl
				<< "end tell" << endl
				<< "end run" << endl;
		}
		else if( CFStringFind( appPath, CFSTR("Entourage"), 0 ).location != kCFNotFound )
		{
			s << "on run argv" << endl
				<< "set vattachment to (item 1 of argv)" << endl
				<< "set vsubject to (item 2 of argv)" << endl
				<< "tell application \"Microsoft Entourage\"" << endl
				<< "set vmessage to make new outgoing message with properties" << endl
				<< "{subject:vsubject, attachments:vattachment}" << endl
				<< "open vmessage" << endl
				<< "activate" << endl
				<< "end tell" << endl
				<< "end run" << endl;
		}
		else if( CFStringFind( appPath, CFSTR("Outlook"), 0 ).location != kCFNotFound )
		{
			s << "on run argv" << endl
				<< "set vattachment to (item 1 of argv)" << endl
				<< "set vsubject to (item 2 of argv)" << endl
				<< "tell application \"Microsoft Outlook\"" << endl
				<< "activate" << endl
				<< "set vmessage to make new outgoing message with properties {subject:vsubject}" << endl
				<< "make new attachment at vmessage with properties {file: vattachment}" << endl
				<< "open vmessage" << endl
				<< "end tell" << endl
				<< "end run" << endl;
		}
#if 0
		else if(CFStringCompare(appPath, CFSTR("/Applications/Thunderbird.app"), 0) == kCFCompareEqualTo)
		{
			// TODO: Handle Thunderbird here? Impossible?
		}
#endif
		CFRelease( appPath );
		p.closeWriteChannel();
		p.waitForFinished();
		if( p.exitCode() == 0 )
			return;
	}
#elif defined(Q_OS_LINUX)
	QByteArray thunderbird;
	QProcess p;
	QStringList env = QProcess::systemEnvironment();
	if( env.indexOf( QRegExp("KDE_FULL_SESSION.*") ) != -1 )
	{
		p.start( "kreadconfig", QStringList()
			<< "--file" << "emaildefaults"
			<< "--group" << "PROFILE_Default"
			<< "--key" << "EmailClient" );
		p.waitForFinished();
		QByteArray data = p.readAllStandardOutput().trimmed();
		if( data.contains("thunderbird") )
			thunderbird = data;
	}
	else if( env.indexOf( QRegExp("GNOME_DESKTOP_SESSION_ID.*") ) != -1 )
	{
		p.start( "gconftool-2", QStringList()
			<< "--get" << "/desktop/gnome/url-handlers/mailto/command" );
		p.waitForFinished();
		QByteArray data = p.readAllStandardOutput();
		if( data.contains("thunderbird") )
			thunderbird = data.split(' ').value(0);
	}
	/*
	else
	{
		p.start( "xprop", QStringList() << "-root" << "_DT_SAVE_MODE" );
		p.waitForFinished();
		if( p.readAllStandardOutput().contains("xfce4") )
		{}
	}*/

	bool status = false;
	if( !thunderbird.isEmpty() )
	{
		status = p.startDetached( thunderbird, QStringList() << "-compose"
			<< QString( "subject='%1',attachment='%2'" )
				.arg( url.queryItemValue( "subject" ) )
				.arg( QUrl::fromLocalFile( url.queryItemValue( "attachment" ) ).toString() ) );
	}
	else
	{
		status = p.startDetached( "xdg-email", QStringList()
			<< "--subject" << url.queryItemValue( "subject" )
			<< "--attach" << url.queryItemValue( "attachment" ) );
	}
	if( status )
		return;
#endif
	QDesktopServices::openUrl( url );
}

QStringList Common::packages( const QStringList &names, bool withName )
{
	QStringList packages;
#if defined(Q_OS_WIN)
	QString path = "SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Uninstall";
#if 1
	int count = applicationOs().contains( "64" ) ? 4 : 2;
	for( int i = 0; i < count; ++i )
	{
		HKEY reg = i % 2 == 0 ? HKEY_CURRENT_USER : HKEY_LOCAL_MACHINE;
		REGSAM param = KEY_READ|(i >= 2 ? KEY_WOW64_32KEY : KEY_WOW64_64KEY);
		HKEY key;
		long result = RegOpenKeyEx( reg, LPCWSTR(path.utf16()), 0, param, &key );
		if( result != ERROR_SUCCESS )
			continue;

		DWORD numSubgroups = 0, maxSubgroupSize = 0;
		result = RegQueryInfoKey( key, 0, 0, 0, &numSubgroups, &maxSubgroupSize, 0, 0, 0, 0, 0, 0 );
		if( result != ERROR_SUCCESS )
		{
			RegCloseKey( key );
			continue;
		}

		for( DWORD j = 0; j < numSubgroups; ++j )
		{
			DWORD groupSize = maxSubgroupSize + 1;
			QString group( groupSize, 0 );
			result = RegEnumKeyEx( key, j, LPWSTR(group.data()), &groupSize, 0, 0, 0, 0 );
			if( result != ERROR_SUCCESS )
				continue;
			group.resize( groupSize );

			HKEY subkey;
			QString subpath = path + "\\" + group;
			result = RegOpenKeyEx( reg, LPCWSTR(subpath.utf16()), 0, param, &subkey );
			if( result != ERROR_SUCCESS )
				continue;

			DWORD numKeys = 0, maxKeySize = 0, maxValueSize = 0;
			result = RegQueryInfoKey( subkey, 0, 0, 0, 0, 0, 0, &numKeys, &maxKeySize, &maxValueSize, 0, 0 );
			if( result != ERROR_SUCCESS )
			{
				RegCloseKey( subkey );
				continue;
			}

			QString name;
			QString version;
			QString type;
			for( DWORD k = 0; k < numKeys; ++k )
			{
				DWORD dataType = 0;
				DWORD keySize = maxKeySize + 1;
				DWORD dataSize = maxValueSize;
				QString key( keySize, 0 );
				QByteArray data( dataSize, 0 );

				result = RegEnumValue( subkey, k, LPWSTR(key.data()), &keySize, 0,
					&dataType, (unsigned char*)data.data(), &dataSize );
				if( result != ERROR_SUCCESS )
					continue;
				key.resize( keySize );
				data.resize( dataSize );

				QString value;
				switch( dataType )
				{
				case REG_SZ:
					value = QString::fromUtf16( (const ushort*)data.constData() );
					break;
				default: continue;
				}

				if( key == "DisplayName" ) name = value;
				if( key == "DisplayVersion" ) version = value;
				if( key == "ReleaseType" ) type = value;
			}
			RegCloseKey( subkey );

			if( !type.contains( "Update", Qt::CaseInsensitive ) &&
				name.contains( QRegExp( names.join( "|" ), Qt::CaseInsensitive ) ) )
				packages << packageName( name, version, withName );
		}
		RegCloseKey( key );
	}
	packages.removeDuplicates();
#else // problems on 64bit windows
	Q_FOREACH( const QString &group, QStringList() << "HKEY_LOCAL_MACHINE" << "HKEY_CURRENT_USER" )
	{
		QSettings s( group + "\\" + path, QSettings::NativeFormat );
		Q_FOREACH( const QString &key, s.childGroups() )
		{
			QString name = s.value( key + "/DisplayName" ).toString();
			QString version = s.value( key + "/DisplayVersion" ).toString();
			QString type = s.value( key + "/ReleaseType" ).toString();
			if( !type.contains( "Update", Qt::CaseInsensitive ) &&
				name.contains( QRegExp( names.join( "|" ), Qt::CaseInsensitive ) ) )
				packages << packageName( name, version, withName );
		}
	}
#endif
#elif defined(Q_OS_MAC)
	QProcess p;

	Q_FOREACH( const QString &name, names )
	{
		p.start( "pkgutil", QStringList() << "--pkg-info-plist" << "ee.sk.idcard." + name );
		p.waitForFinished();

		QString result;
		QXmlStreamReader xml( &p );
		while( xml.readNext() != QXmlStreamReader::Invalid )
		{
			if( !xml.isStartElement() || xml.name() != "key" || xml.readElementText() != "pkg-version" )
				continue;
			xml.readNextStartElement();
			result = packageName( name, xml.readElementText(), withName );
			break;
		}
		if( !result.isEmpty() )
		{
			packages << result;
			continue;
		}

		QStringList params = QStringList() << "read";
		if( QFile::exists( "/Applications/" + name + ".app/Contents/Info.plist" ) )
			params << "/Applications/" + name + ".app/Contents/Info" << "CFBundleShortVersionString";
		else if( QFile::exists( "/var/db/receipts/ee.sk.idcard." + name + ".plist" ) )
			params << "/var/db/receipts/ee.sk.idcard." + name << "PackageVersion";
		else if( QFile::exists( "/Library/Receipts/" + name + ".pkg/Contents/Info.plist" ) )
			params << "/Library/Receipts/" + name + ".pkg/Contents/Info" << "CFBundleShortVersionString";
		else
			continue;

		p.start( "defaults", params );
		p.waitForFinished();
		packages << packageName( name, QString::fromLocal8Bit( p.readAll().trimmed() ), withName );
	}
#elif defined(Q_OS_LINUX)
	QProcess p;

	Q_FOREACH( const QString &name, names )
	{
		p.start( "dpkg-query", QStringList() << "-W" << "-f=${Version}" << name );
		if( !p.waitForStarted() && p.error() == QProcess::FailedToStart )
		{
			p.start( "rpm", QStringList() << "-q" << "--qf" << "%{VERSION}" << name );
			p.waitForStarted();
		}
		p.waitForFinished();
		if( !p.exitCode() )
			packages << packageName( name, QString::fromLocal8Bit( p.readAll().trimmed() ), withName );
	}
#endif
	return packages;
}

bool Common::runPrivileged( const QString &program, const QStringList &arguments )
{
#if defined(Q_OS_WIN)
	QString params = arguments.join(" ");
	HINSTANCE result = ShellExecute( 0, L"runas", (const wchar_t*)program.utf16(), (const wchar_t*)params.utf16(), 0, SW_HIDE );
	return int(result) > 32;
#elif defined(Q_OS_MAC) && defined(USE_SECURITY)
	QList<QByteArray> u8args;
	Q_FOREACH( const QString arg, arguments )
		u8args << arg.toUtf8();

	std::vector<char*> args;
	for( int i = 0; i < u8args.size(); ++i )
		args.push_back( (char*)u8args[i].constData() );
	args[u8args.size()] = 0;
	FILE *pipe = 0;

	AuthorizationRef ref;
	AuthorizationCreate( 0, kAuthorizationEmptyEnvironment, kAuthorizationFlagDefaults, &ref );
	OSStatus status = AuthorizationExecuteWithPrivileges( ref, program.toUtf8().constData(),
		kAuthorizationFlagDefaults, args.data(), &pipe );

	if( pipe )
	{
		QFile log;
		log.open( pipe, QFile::ReadOnly );
		qWarning() << log.readAll();
		log.close();
		fclose( pipe );
	}

	AuthorizationFree( ref, kAuthorizationFlagDestroyRights );
	return status == errAuthorizationSuccess;
#else
	return QProcess::startDetached( program, arguments );
#endif
	return false;
}

void Common::setAccessibleName( QLabel *widget )
{
	QTextDocument doc;
	doc.setHtml( widget->text() );
	widget->setAccessibleName( doc.toPlainText() );
}

void Common::showHelp( const QString &msg, int code )
{
	QUrl u;

	if( code > 0 )
	{
		u.setUrl( "http://www.sk.ee/digidoc/support/errorinfo/" );
		u.addQueryItem( "app", applicationName() );
		u.addQueryItem( "appver", applicationVersion() );
		u.addQueryItem( "l", Settings::language() );
		u.addQueryItem( "code", QString::number( code ) );
		u.addQueryItem( "os", applicationOs() );
	}
	else
	{
		u = helpUrl();
		u.addQueryItem( "searchquery", msg );
		u.addQueryItem( "searchtype", "all" );
		u.addQueryItem( "_m", "core" );
		u.addQueryItem( "_a", "searchclient" );
	}
	QDesktopServices::openUrl( u );
}

bool Common::startDetached( const QString &program, const QStringList &arguments )
{
#ifdef Q_OS_MAC
	return QProcess::startDetached( "/usr/bin/open", QStringList() << "-a" << program << arguments );
#else
	return QProcess::startDetached( program, arguments );
#endif
}

QString Common::tempFilename()
{
	QString path = QDir::tempPath();
	QString prefix = QFileInfo( applicationFilePath() ).baseName();
#ifdef Q_OS_WIN
	wchar_t *name = _wtempnam( (wchar_t*)QDir::toNativeSeparators( path ).utf16(), (wchar_t*)prefix.utf16() );
	QString result = QString::fromUtf16( (ushort*)name );
#else
	char *name = tempnam( path.toLocal8Bit(), prefix.toLocal8Bit() );
	QString result = QString::fromLocal8Bit( name );
#endif
	free( name );
	return result;
}
