/*
 * QEstEidCommon
 *
 * Copyright (C) 2009-2012 Jargo Kõster <jargo@innovaatik.ee>
 * Copyright (C) 2009-2012 Raul Metsma <raul@innovaatik.ee>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include "Diagnostics.h"

#include "Common.h"
#include "QPCSC.h"

#include <QtCore/QFile>
#include <QtCore/QTextStream>
#include <QtCore/QProcess>

#include <sys/utsname.h>

#ifndef APPSTORE
static QByteArray runProcess( const QString &program, const QStringList &arguments = QStringList() )
{
	QProcess p;
	p.start( program, arguments );
	p.waitForFinished();
	return p.readAll();
}
#endif

void Diagnostics::run()
{
	QString info;
	QTextStream s( &info );

	s << "<b>" << tr("Locale (time-, number format / codepage):") << "</b> ";
	QLocale::Language language = QLocale::system().language();
	QString locale = QString( "%1 / ").arg( language == QLocale::C ? "English/United States" : QLocale::languageToString( language ) );
	QString ctype = QProcessEnvironment::systemEnvironment().value( "LC_CTYPE" );
	locale += !ctype.isEmpty() ? ctype : QProcessEnvironment::systemEnvironment().value( "LANG" );
	s << locale << "<br /><br />";
	emit update( info );
	info.clear();

#ifndef APPSTORE
	QStringList package = Common::packages( QStringList() << "estonianidcard", false );
	if( !package.isEmpty() )
		s << "<b>" << tr("Base version:") << "</b> " << package.first() << "<br />";
#endif
	s << "<b>" << tr("Application version:") << "</b> " << QCoreApplication::applicationVersion() << "<br />";
	emit update( info );
	info.clear();

	s << "<b>" << tr("OS:") << "</b> " << Common::applicationOs() << "<br />";
#ifdef Q_OS_LINUX
	s << "<b>" << tr("CPU:") << "</b> ";
	QFile f( "/proc/cpuinfo" );
	if( f.open( QFile::ReadOnly ) )
	{
		QRegExp rx( "model name.*\\: (.*)\n" );
		rx.setMinimal( true );
		int pos = rx.indexIn( QString::fromLocal8Bit( f.readAll() ) );
		if( pos != -1 )
			s << rx.cap( 1 );
	}
#elif !defined(APPSTORE)
	s << "<b>" << tr("CPU:") << "</b> ";
	QString result = runProcess( "system_profiler", QStringList() << "SPHardwareDataType" );
	QRegExp reg( "Processor Name:(\\s*)(.*)\n.*Processor Speed:(\\s*)(.*)\n" );
	reg.setMinimal( true );
	s << (reg.indexIn( result ) != -1 ? reg.cap( 2 ) + " (" + reg.cap( 4 ) + ")" : "");
#endif
	s << "<br />";
	struct utsname unameData;
	uname(&unameData);
	s << "<b>" << tr("Kernel:") << "</b> "
	  << unameData.sysname << " " << unameData.release << " "
	  << unameData.version << " " << unameData.machine << "<br /><br />";
	emit update( info );
	info.clear();

	s << "<b>" << tr("Arguments:") << "</b> " << qApp->arguments().join(" ") << "<br />";
	s << "<b>" << tr("Library paths:") << "</b> " << QCoreApplication::libraryPaths().join( ";" ) << "<br />";
	s << "<b>" << tr("Libraries") << ":</b><br />";
#ifndef APPSTORE
	s << Common::packages( QStringList() << "libdigidoc" << "libdigidocpp" << "qdigidoc" << "qesteidutil" ).join( "<br />" ) << "<br />";
#ifdef Q_OS_MAC
	s << runProcess(  "/Library/OpenSC/bin/opensc-tool", QStringList() << "-i" ) << "<br />";
#else
	s << Common::packages( QStringList() << "openssl" << "libpcsclite1" << "pcsc-lite" << "opensc" ).join( "<br />" ) << "<br />";
#endif
#endif
	s << "QT (" << qVersion() << ")<br />" << "<br />";
	emit update( info );
	info.clear();

	s << getReaderInfo() << "<br /><br />";
	emit update( info );
	info.clear();

#ifndef APPSTORE
	QStringList browsers = Common::packages( QStringList()
#ifdef Q_OS_MAC
		<< "Google Chrome" << "Firefox" << "Safari" );
#else
		<< "chromium-browser" << "firefox" << "MozillaFirefox" );
#endif
	if( !browsers.isEmpty() )
		s << "<b>" << tr("Browsers:") << "</b><br />" << browsers.join( "<br />" ) << "<br /><br />";
	emit update( info );
	info.clear();

	QString cmd = QString::fromUtf8( runProcess( "opensc-tool", QStringList() << "-la" ) );
	if( !cmd.isEmpty() )
		s << "<b>" << tr("OpenSC tool:") << "</b><br/> " << cmd.replace( "\n", "<br />" ) << "<br />";
	emit update( info );
	info.clear();

	cmd = QString::fromUtf8( runProcess( "pkcs11-tool",
		QStringList() << "-T" << QString("--module=%1").arg( PKCS11_MODULE ) ) );
	if( !cmd.isEmpty() )
		s << "<b>" << tr("PKCS11 tool:") << "</b><br/> " << cmd.replace( "\n", "<br />" ) << "<br />";
	emit update( info );
	info.clear();

#ifdef Q_OS_LINUX
	cmd = runProcess( "lsusb" );
#else
	cmd = runProcess( "system_profiler", QStringList() << "SPUSBDataType" );
#endif
	if( !cmd.isEmpty() )
		s << "<b>" << tr("USB info:") << "</b><br/> " << cmd.replace( "\n", "<br />" ) << "<br />";
	emit update( info );
	info.clear();
#endif

	QApplication::restoreOverrideCursor();
}
