/*
 * QEstEidUtil
 *
 * Copyright (C) 2011-2012 Jargo Kõster <jargo@innovaatik.ee>
 * Copyright (C) 2011-2012 Raul Metsma <raul@innovaatik.ee>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include "QSmartCard_p.h"

#include <common/Common.h>
#include <common/IKValidator.h>
#include <common/PinDialog.h>
#include <common/Settings.h>

#include <QtCore/QDateTime>
#include <QtCore/QDebug>
#include <QtCore/QDir>
#include <QtCore/QProcess>
#include <QtCore/QScopedPointer>
#include <QtNetwork/QSslCertificate>
#include <QtNetwork/QSslKey>

#include <openssl/evp.h>

QSmartCardData::QSmartCardData(): d( new QSmartCardDataPrivate ) {}
QSmartCardData::QSmartCardData( const QSmartCardData &other ): d( other.d ) {}
QSmartCardData::~QSmartCardData() {}
QSmartCardData QSmartCardData::operator =( const QSmartCardData &other ) { d = other.d; return *this; }

QString QSmartCardData::card() const { return d->card; }
QStringList QSmartCardData::cards() const { return d->cards; }

bool QSmartCardData::isPinpad() const { return d->pinpad; }
bool QSmartCardData::isSecurePinpad() const
{ return d->reader.contains( "EZIO SHIELD", Qt::CaseInsensitive ); }
bool QSmartCardData::isValid() const
{ return d->data.value( Expiry ).toDateTime() >= QDateTime::currentDateTime(); }

QString QSmartCardData::reader() const { return d->reader; }
QStringList QSmartCardData::readers() const { return d->readers; }

QVariant QSmartCardData::data( PersonalDataType type ) const
{ return d->data.value( type ); }
SslCertificate QSmartCardData::authCert() const { return d->authCert; }
SslCertificate QSmartCardData::signCert() const { return d->signCert; }

quint8 QSmartCardData::retryCount( PinType type ) const
{
	switch( type )
	{
	case Pin1Type: return d->pin1retry;
	case Pin2Type: return d->pin2retry;
	case PukType: return d->pukretry;
	}
	return -1;
}

ulong QSmartCardData::usageCount( PinType type ) const
{
	switch( type )
	{
	case Pin1Type: return d->pin1usage;
	case Pin2Type: return d->pin2usage;
	default: return 0;
	}
}

QSmartCardData::CardVersion QSmartCardData::version() const { return d->version; }

QString QSmartCardData::typeString( QSmartCardData::PinType type )
{
	switch( type )
	{
	case Pin1Type: return "PIN1";
	case Pin2Type: return "PIN2";
	case PukType: return "PUK";
	}
	return "";
}



int QSmartCardPrivate::rsa_sign( int type, const unsigned char *m, unsigned int m_len,
		unsigned char *sigret, unsigned int *siglen, const RSA *rsa )
{
	if( type != NID_md5_sha1 && m_len != 36 )
		return 0;

	QSmartCardPrivate *d = (QSmartCardPrivate*)RSA_get_app_data( rsa );
	if ( !d )
		return 0;

	try
	{
		ByteVec vec = d->card->calcSSL( ByteVec( m, m + m_len ) );
		if( vec.size() == 0 )
			return 0;
		*siglen = (unsigned int)vec.size();
		qMemCopy( sigret, &vec[0], vec.size() );
		return 1;
	}
	catch( const std::runtime_error &e )
	{
		qDebug() << Q_FUNC_INFO << e.what();
	}
	return 0;
}

QSmartCard::ErrorType QSmartCardPrivate::handleAuthError( QSmartCardData::PinType type, const AuthError &e )
{
	updateCounters( t.d );
	if( t.retryCount( type ) < 1 ) return QSmartCard::BlockedError;
	switch( e.SW1 )
	{
	case 0x63:
		switch( e.SW2 )
		{
		case 0xC0: return QSmartCard::BlockedError; //pin retry count 0
		case 0xC1: // Validate error, 1 tries left
		case 0xC2: // Validate error, 2 tries left
		default: return QSmartCard::ValidateError;
		}
	case 0x64:
		switch( e.SW2 )
		{
		case 0x00: // Timeout (SCM)
		case 0x01: return QSmartCard::CancelError; // Cancel (OK, SCM)
		case 0x02: return QSmartCard::DifferentError;
		case 0x03: return QSmartCard::LenghtError;
		default: break;
		}
		break;
	case 0x69:
		if( e.SW2 == 0x83 ) return QSmartCard::BlockedError;
		if( e.SW2 == 0x85 ) return QSmartCard::OldNewPinSameError;
		break;
	case 0x6A:
		if( e.SW2 == 0x80 ) return QSmartCard::OldNewPinSameError;
		break;
	default:
		break;
	}
	return QSmartCard::UnknownError;
}

void QSmartCardPrivate::updateCounters( QSmartCardDataPrivate *d )
{
	try
	{
		d->pukretry = d->pin1retry = d->pin2retry = -1;
		d->pin1usage = d->pin2usage = 0;
		card->getRetryCounts( d->pukretry, d->pin1retry, d->pin2retry );
		card->getKeyUsageCounters( d->pin1usage, d->pin2usage );
	}
	catch( const std::runtime_error &e )
	{
		qDebug() << Q_FUNC_INFO << e.what();
	}
}




QSmartCard::QSmartCard( const QString &lang, QObject *parent )
:	QThread( parent )
,	d( new QSmartCardPrivate )
{
	setLang( lang );
	d->t.d->readers = QStringList() << "loading";
	d->t.d->cards = QStringList() << "loading";
	d->t.d->card = "loading";

	if( QProcessEnvironment::systemEnvironment().contains( "QESTEIDUTIL_DEBUG" ) )
	{
		QString path = QDir::tempPath() + "/qesteidutil.log";
#ifdef Q_OS_WIN
		d->log.open( (wchar_t*)path.utf16() );
#else
		d->log.open( path.toLocal8Bit() );
#endif
	}
}

QSmartCard::~QSmartCard()
{
	d->terminate = true;
	wait();
	delete d;
}

QSmartCard::ErrorType QSmartCard::change( QSmartCardData::PinType type, const QString &newpin, const QString &pin )
{
	QMutexLocker locker( &d->m );
	if( !d->card ) return QSmartCard::UnknownError;
	d->cmd = QSmartCardPrivate::Change;
	d->type = type;
	d->pin = newpin.toUtf8().constData();
	d->old = pin.toUtf8().constData();
	d->e.exec();
	return d->result;
}

QSmartCardData QSmartCard::data() const { return d->t; }

Qt::HANDLE QSmartCard::key()
{
	RSA *rsa = RSAPublicKey_dup( (RSA*)d->t.authCert().publicKey().handle() );
	if ( !rsa )
		return 0;

	RSA_set_method( rsa, &d->method );
	rsa->flags |= RSA_FLAG_SIGN_VER;
	RSA_set_app_data( rsa, d );
	EVP_PKEY *key = EVP_PKEY_new();
	EVP_PKEY_set1_RSA( key, rsa );
	RSA_free( rsa );
	return Qt::HANDLE(key);
}

QSmartCard::ErrorType QSmartCard::login( QSmartCardData::PinType type )
{
	PinDialog::PinFlags flags = PinDialog::Pin1Type;
	QSslCertificate cert;
	switch( type )
	{
	case QSmartCardData::Pin1Type: flags = PinDialog::Pin1Type; cert = d->t.authCert(); break;
	case QSmartCardData::Pin2Type: flags = PinDialog::Pin2Type; cert = d->t.signCert(); break;
	default: return UnknownError;
	}

	QScopedPointer<PinDialog> p;
	PinString pin;
	if( !d->t.isPinpad() )
	{
		p.reset( new PinDialog( flags, cert, 0, qApp->activeWindow() ) );
		if( !p->exec() )
			return CancelError;
		pin = p->text().toUtf8().constData();
	} else {
		p.reset( new PinDialog( PinDialog::PinFlags(flags|PinDialog::PinpadFlag), cert, 0, qApp->activeWindow() ) );
		p->open();
	}

	d->m.lock();
	if( !d->card ) return QSmartCard::UnknownError;
	d->cmd = QSmartCardPrivate::Validate;
	d->type = type;
	d->pin = pin;
	connect( this, SIGNAL(eventStarted()), p.data(), SIGNAL(startTimer()) );
	d->e.exec();
	if( d->result != QSmartCard::NoError )
		d->m.unlock();
	return d->result;
}

void QSmartCard::logout()
{
	d->updateCounters( d->t.d );
	d->m.unlock();
}

void QSmartCard::reload() { selectCard( d->t.card() ); }

void QSmartCard::run()
{
	while( !d->terminate )
	{
		if( d->m.tryLock() )
		{
			try
			{
				if( d->numReaders == 0 ) // to force cache reload
				{
					delete d->card;
					delete d->manager;
					d->card = 0;
					d->manager = 0;
				}
				d->numReaders = 0;

				if( !d->manager )
				{
					d->manager = new PCSCManager();
					d->manager->setReaderLanguageId( d->lang );
				}
				if( d->log.is_open() )
					d->manager->setLogging( &d->log );

				d->numReaders = d->manager->getReaderCount( true );
			}
			catch( const std::runtime_error &e )
			{
				qDebug() << Q_FUNC_INFO << e.what() << d->numReaders;
			}

			// Get list of available cards
			QMap<QString,quint8> cards;
			QStringList readers;
			for( quint8 i = 0; i < d->numReaders; ++i )
			{
				try
				{
					std::string name = d->manager->getReaderName(i);
					std::string state = d->manager->getReaderState(i);
					qDebug() << Q_FUNC_INFO << name.c_str() << state.c_str();

					readers << QString::fromUtf8( name.c_str() );
					if( state.find( "PRESENT" ) == std::string::npos )
						continue;

					EstEidCard card( *d->manager, i );
					if( !card.isInReader(i) )
						continue;

					QString cardid = QString::fromStdString( card.readDocumentID() );
					cards[cardid] = i;
					qDebug() << Q_FUNC_INFO << cardid;
				}
				catch( const std::runtime_error &e )
				{
					qDebug() << Q_FUNC_INFO << e.what();
				}
			}

			// cardlist has changed
			QStringList order = cards.keys();
			std::sort( order.begin(), order.end(), Common::cardsOrder );
			bool update = d->t.cards() != order || d->t.readers() != readers;

			// check if selected card is still in slot
			if( !d->t.card().isEmpty() && !order.contains( d->t.card() ) )
			{
				update = true;
				d->t.d = new QSmartCardDataPrivate();
			}

			d->t.d->cards = order;
			d->t.d->readers = readers;


			// if none is selected select first from cardlist
			if( d->t.card().isEmpty() && !d->t.cards().isEmpty() )
				selectCard( d->t.cards().first() );

			// read card data
			if( d->t.cards().contains( d->t.card() ) && (!d->card || d->t.authCert().isNull() || d->t.signCert().isNull()) )
			{
				update = true;
				try
				{
					delete d->card;
					d->card = new EstEidCard( *d->manager, cards.value( d->t.card() ) );
					d->card->isInReader( cards.value( d->t.card() ) );

					QSmartCardDataPrivate *t = d->t.d;
					d->updateCounters( t );
					std::vector<std::string> data;
					d->card->readPersonalData( data, EstEidCard::SURNAME, EstEidCard::COMMENT4 );
					ByteVec authcert = d->card->getAuthCert();
					ByteVec signcert = d->card->getSignCert();

					t->reader = readers.value( cards.value( d->t.card() ) );
					t->pinpad = d->card->isSecureConnection();
					t->version = QSmartCardData::CardVersion(d->card->getCardVersion());
					t->data[QSmartCardData::SurName] =
						SslCertificate::formatName( d->encode( data[EstEidCard::SURNAME] ) ).trimmed();
					t->data[QSmartCardData::FirstName] =
						SslCertificate::formatName( d->encode( data[EstEidCard::FIRSTNAME] ) ).trimmed();
					t->data[QSmartCardData::MiddleName] =
						SslCertificate::formatName( d->encode( data[EstEidCard::MIDDLENAME] ) ).trimmed();
					t->data[QSmartCardData::Sex] = d->encode( data[EstEidCard::SEX] );
					t->data[QSmartCardData::Citizen] = d->encode( data[EstEidCard::CITIZEN] );
					t->data[QSmartCardData::BirthDate] = QDate::fromString( d->encode( data[EstEidCard::BIRTHDATE] ), "dd.MM.yyyy" );
					t->data[QSmartCardData::Id] = d->encode( data[EstEidCard::ID] );
					t->data[QSmartCardData::DocumentId] = d->encode( data[EstEidCard::DOCUMENTID] );
					t->data[QSmartCardData::Expiry] = QDate::fromString( d->encode( data[EstEidCard::EXPIRY] ), "dd.MM.yyyy" );
					t->data[QSmartCardData::BirthPlace] =
						SslCertificate::formatName( d->encode( data[EstEidCard::BIRTHPLACE] ) );
					t->data[QSmartCardData::IssueDate] = QDate::fromString( d->encode( data[EstEidCard::ISSUEDATE] ), "dd.MM.yyyy" );
					t->data[QSmartCardData::ResidencePermit] = d->encode( data[EstEidCard::RESIDENCEPERMIT] );
					t->data[QSmartCardData::Comment1] = d->encode( data[EstEidCard::COMMENT1] );
					t->data[QSmartCardData::Comment2] = d->encode( data[EstEidCard::COMMENT2] );
					t->data[QSmartCardData::Comment3] = d->encode( data[EstEidCard::COMMENT3] );
					t->data[QSmartCardData::Comment4] = d->encode( data[EstEidCard::COMMENT4] );

					if( !authcert.empty() )
						t->authCert = QSslCertificate( QByteArray( (char*)&authcert[0], (int)authcert.size() ), QSsl::Der );
					if( !signcert.empty() )
						t->signCert = QSslCertificate( QByteArray( (char*)&signcert[0], (int)signcert.size() ), QSsl::Der );

					QStringList mailaddresses = t->authCert.alternateSubjectNames().values( QSsl::EmailEntry );
					t->data[QSmartCardData::Email] = !mailaddresses.isEmpty() ? mailaddresses.first() : "";

					QStringList name = QStringList()
						<< t->data[QSmartCardData::FirstName].toString()
						<< t->data[QSmartCardData::MiddleName].toString()
						<< t->data[QSmartCardData::SurName].toString();
					name.removeAll( "" );
					t->data[QSmartCardData::FullName] = name.join( " " );

					switch( t->authCert.type() )
					{
					case SslCertificate::DigiIDType:
					case SslCertificate::DigiIDTestType:
						t->data[QSmartCardData::FullName] = t->authCert.toString( "GN SN" );
						t->data[QSmartCardData::Id] = t->authCert.subjectInfo("serialNumber");
						t->data[QSmartCardData::BirthDate] = IKValidator::birthDate( t->authCert.subjectInfo("serialNumber") );
						t->data[QSmartCardData::IssueDate] = t->authCert.effectiveDate();
						t->data[QSmartCardData::Expiry] = t->authCert.expiryDate();
						break;
					case SslCertificate::TempelType:
					case SslCertificate::TempelTestType:
						t->data[QSmartCardData::FullName] = t->authCert.subjectInfo( QSslCertificate::CommonName );
						t->data[QSmartCardData::Email] = t->authCert.subjectInfo( "emailAddress" );
						t->data[QSmartCardData::Id] = t->authCert.subjectInfo("serialNumber");
						t->data[QSmartCardData::IssueDate] = t->authCert.effectiveDate();
						t->data[QSmartCardData::Expiry] = t->authCert.expiryDate();
						break;
					default: break;
					}
				}
				catch( std::runtime_error &e )
				{
					qDebug() << Q_FUNC_INFO << "Error on loading card data: " << e.what();
				}
			}

			// update data if something has changed
			if( update )
				Q_EMIT dataChanged();
			d->m.unlock();
		}
		else if( d->e.isRunning() )
		{
			emit eventStarted();
			try
			{
				byte retries = 0;
				d->result = QSmartCard::UnknownError;
				switch( d->cmd )
				{
				case QSmartCardPrivate::Change:
					d->card->reconnectWithT0();
					switch( d->type )
					{
					case QSmartCardData::Pin1Type:
						d->card->changeAuthPin( d->pin, d->old, retries );
						break;
					case QSmartCardData::Pin2Type:
						d->card->changeSignPin( d->pin, d->old, retries );
						break;
					case QSmartCardData::PukType:
						d->card->changePUK( d->pin, d->old, retries );
						break;
					}
					d->updateCounters( d->t.d );
					break;
				case QSmartCardPrivate::Unblock:
					d->card->reconnectWithT0();
					switch( d->type )
					{
					case QSmartCardData::Pin1Type:
						d->card->unblockAuthPin( d->pin, d->old, retries );
						break;
					case QSmartCardData::Pin2Type:
						d->card->unblockSignPin( d->pin, d->old, retries );
						break;
					default: break;
					}
					d->updateCounters( d->t.d );
					break;
				case QSmartCardPrivate::Validate:
					d->card->reconnectWithT0();
					switch( d->type )
					{
					case QSmartCardData::Pin1Type:
						d->card->validateAuthPin( d->pin, retries );
						break;
					case QSmartCardData::Pin2Type:
						d->card->validateSignPin( d->pin, retries );
						break;
					case QSmartCardData::PukType:
						d->card->validatePuk( d->pin, retries );
						break;
					default: break;
					}
					break;
				case QSmartCardPrivate::ValidateInternal:
					// Dont reconnect, xp fails to update certificates
					switch( d->type )
					{
					case QSmartCardData::Pin1Type:
						d->card->enterPin( EstEidCard::PIN_AUTH, d->pin );
						break;
					case QSmartCardData::Pin2Type:
						d->card->enterPin( EstEidCard::PIN_SIGN, d->pin );
						break;
					case QSmartCardData::PukType:
						d->card->enterPin( EstEidCard::PUK, d->pin );
						break;
					}
					break;
				default: break;
				}
				d->result = QSmartCard::NoError;
			}
			catch( const AuthError &e )
			{
				d->result = d->handleAuthError(
					d->cmd == QSmartCardPrivate::Unblock ? QSmartCardData::PukType : d->type, e );
			}
			catch( const std::runtime_error &e )
			{
				qDebug() << Q_FUNC_INFO << e.what();
			}
			d->pin.clear();
			d->old.clear();
			d->e.quit();
		}

		sleep( 5 );
	}
}

void QSmartCard::selectCard( const QString &card )
{
	QSmartCardDataPrivate *t = new QSmartCardDataPrivate();
	t->cards = d->t.cards();
	t->readers = d->t.readers();
	t->card = card;
	d->t.d = t;
	Q_EMIT dataChanged();
}

void QSmartCard::setLang( const QString &lang )
{
	d->lang = EstEidCard::EST;
	if( lang == "en" ) d->lang = EstEidCard::ENG;
	if( lang == "ru" ) d->lang = EstEidCard::RUS;
	if( d->manager )
		d->manager->setReaderLanguageId( d->lang );
}

QSmartCard::ErrorType QSmartCard::unblock( QSmartCardData::PinType type, const QString &pin, const QString &puk )
{
	QMutexLocker locker( &d->m );
	if( !d->card ) return QSmartCard::UnknownError;
	d->cmd = QSmartCardPrivate::Unblock;
	d->type = type;
	d->pin = pin.toUtf8().constData();
	d->old = puk.toUtf8().constData();
	d->e.exec();
	return d->result;
}
